#
# Author: Justin Engler
#
# Copyright (c) 2011 RAFT Team
#
# This file is part of RAFT.
#
# RAFT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# RAFT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RAFT.  If not, see <http://www.gnu.org/licenses/>.
#
import json
import ast
from utility import TreeWidgetTools
from PyQt4.QtGui import *
from PyQt4.QtCore import Qt
import pprint


class AbstractAnalyzer(object):
    """
    AbstractAnalyzer
    Defines the interface for a RAFT analyzer.
    
    Analyzers search through request or response data,
    locate items of interest, and report back the details
    
    Analyzers shall either inherit from AbstractAnalyzer
    or implement all of the member functions with compatible
    signatures and return values.
    
    Each analyzer will be instantiated before analysis begins, and each
    resulting analysis object will be reused on each page.
    It is the responsibility of the analyzer writer to save or clear
    object fields as desired for each page.
    """
    
    #####################Standard Functions
    #####################Override when defining your own analyzers
    def __init__(self):
        self.desc="Did you define a desc in your subclass? "\
              "An abstract class to define what an analyzer must support."\
              "  Automatically called during analysis or configuration start."
        self.friendlyname="Abstract Analyzer"
        self.currentconfiguration=None    
        
        
        
    def preanalysis(self):
        """Any special pre-test setup goes here.  This will be called after configuration, before any pages are loaded
        """
        pass
    
    def analyzeTransaction(self,transaction,results):
        """
        Performs an analysis on target. This function must be defined by a subclass.
        Transaction is information about the request and response (the target of your analysis)
        """
        raise Exception("analyze() must be defined in a subclass for analysis")
    
    
    def postanalysis(self,results):
        """Any post-test steps go here.  Called after all pages were analyzed."""
        pass
    
    def getDefaultConfiguration(self):
        """Returns the default configuration dictionary for this analyzer.
           If your module is configurable at all, you must override this."""
        return {}
    
    #####################Special Functions
    #####################You shouldn't need to override these unless you're doing something very special
        
    def setConfiguration(self,newconfiguration):
        """Accepts any configuration data from the system.
            Either a settings dictionary or a JSON string representation
            of the same is OK."""
        if type(newconfiguration)==str or type(newconfiguration)==unicode:
            #print "FOUND JSON, decoding"
            #pprint.pprint(newconfiguration)
            if newconfiguration:
                temp=json.loads(newconfiguration)
                self.currentconfiguration = temp
            
        elif type(newconfiguration)==dict:
            #print "FOUND DICT"
            self.currentconfiguration = newconfiguration
        else:
            raise ValueError("I don't know what that configuration is.")
        #pprint.pprint(self.currentconfiguration)
        return self.currentconfiguration
    
    def encodeConfiguration(self,newconfiguration):
        """Takes a given configuration and encodes it for storage.
            Does not alter existing runtime config of this analyzer.
        """
        return json.dumps(newconfiguration)
        
        
        
    def initResultsData(self):
        """Sets up results data.  Called after preanalysis, but before actual analysis.
        Analyzer writers shouldn't need to override this function except for special cases, like
        when subclassing a custom AnalysisResults object.
        """
        self.analysisresults=AnalysisResults()
        self.analysisresults.setAnalyzerInfo(self.desc,self.friendlyname,str(self.__class__).translate(None,'<>'))
        
        
    def getResults(self):
        """returns a reference to the Analyzer's results collection.
        Analyzer writers shouldn't need to override this function"""
        return self.analysisresults
    
    
    def getCurrentConfiguration(self):
        """Returns the current config dict.  Should not need to override this"""
        
        if not hasattr(self,'currentconfiguration') or self.currentconfiguration is None:
            self.currentconfiguration=self.getDefaultConfiguration()
            #print 'Using Default Config'
        return self.currentconfiguration
        
        
    def generateConfigurationGui(self):
        currentconfig=self.getCurrentConfiguration()
        return TreeWidgetTools.generate_tree_widget(currentconfig)
        
        
    def generalInfoToHTML(self):
        """Returns an HTML 'header' string describing the analyzer"""
        outstring="""<h1>%s</h1>
        (%s)
        <p>%s</p>
        """%(self.friendlyname,str(self.__class__).translate(None,'<>'),self.desc)
        return outstring
    
#########################################################################################################
#TODO: Break classes below out into one or more new files.
#########################################################################################################

class AnalysisResults(object):
    """Contains all results found for a given analysis."""
    
    def __init__(self):
        """Results that span across multiple pages"""
        self.overall={}
        
        """Results that apply to a single page"""
        self.pages={} 
        
        """pages scanned with no results"""
        self.nofindings={}
        
        """counts of total results within each grouping"""
        self.resultcounts={'Overall':0,'Page':0,'No':0}
        
        self.desc=None
        self.friendlyname=None
        self.analyzerclass=None
        
        
        
#############################Standard Functions
#############################Functions often used when writing an analyzer
    def addPageResult(self, pageid, url, type, desc, data, span=None, severity=None, certainty=None):
            """Adds a new per-page standard result to the given pageid for this analysis"""
            self.addCustomPageResult(pageid,SingleResult(type, desc, data, span, severity, certainty),url)
            
    def addOverallResult(self, type, desc, data, span=None, severity=None, certainty=None, context=None):
            """Adds a new overall result to this analysis"""
            self.addCustomOverallResult(SingleResult(type, desc, data, span, severity, certainty),context)
            
#############################Special Functions
#############################You shouldn't need to call these unless you're doing something crazy.
    def addCustomPageResult(self,pageid,result,url):
        if pageid not in self.pages:
            self.pages[pageid]=ResultSet(pageid,False,url)
        self.pages[pageid].addResult(result)
        
    def addCustomOverallResult(self,result,context):
        """Adds an arbitrary result object to the overall results."""
        if context not in self.overall:
            self.overall[context]=ResultSet(None,True,context)
        self.overall[context].addResult(result)
        
    def setAnalyzerInfo(self, newdesc,newfriendlyname, newanalyzerclass):
        self.desc=newdesc
        self.friendlyname=newfriendlyname
        self.analyzerclass=newanalyzerclass
        
        
    def toHTML(self):
        """returns an HTML representation of the entire analysis"""
        finaloutput=self.generalInfoToHTML()
        
        if len(self.overall) > 0:
            finaloutput+='<h2>Overall Results</h2>'
            for k in self.overall.keys():
                finaloutput+=self.overall[k].toHTML()
                
               
        if len(self.pages)>0:
            finaloutput+='<h2>Results for each page analyzed</h2>'
            for k in self.pages.keys():
                finaloutput+=self.pages[k].toHTML()
        return finaloutput
        
    def generalInfoToHTML(self):
        """Returns an HTML 'header' string describing the test performed"""
        outstring="""<h1>%s</h1>
                    <p>(%s)</p>
                    <p>%s</p>
                    """%(self.friendlyname,self.analyzerclass,self.desc)
        return outstring
        
    def generateTreeItem(self,parentnode):
        tempitem=QTreeWidgetItem(parentnode)
        tempitem.setText(0,str(self.friendlyname))
        tempitem.setText(1,"".join((str(self.numresults),' results')))
        tempitem.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
        tempitem.customdata=self
        return tempitem
    
    def generateTreeChildren(self,db,cursor,parentnode):
        #If this tree item came from the db, and we haven't populated it yet, populate it.
        if self.dbgenerated and not self.dbretrieved:
            resultsets=db.analysis_get_resultsets_per_instance(cursor,self.instanceid)
            for resultset in resultsets:
                numresults=resultset[4]
                #AnalysisResultSet_ID, Response_analyzed, isOverallResult, context
                if resultset[2]:
                    store=self.overall
                    storekey=resultset[3]
                    self.resultcounts['Overall']+=numresults
                    tempRS=ResultSet(storekey,None,True)

                elif numresults>0:
                    store=self.pages
                    storekey=resultset[1]
                    self.resultcounts['Page']+=numresults
                    tempRS=ResultSet(resultset[3],storekey,False)
                
                else:
                    store=self.nofindings
                    storekey=resultset[1]
                    tempRS=ResultSet(resultset[3],storekey,False)
                
                    
                tempRS.dbgenerated=True
                tempRS.dbretrieved=False
                tempRS.resultsetid=resultset[0]
                tempRS.numresults=resultset[4]
                store[storekey]=tempRS
            self.dbretrieved=True
                
        #Now that the tree is populated, make the nodes
        childnodes=list()
        for name,store in (('Overall',self.overall),('Page',self.pages), ('No',self.nofindings)):
            storelen=len(store)
            if storelen>0:
                tempitem=QTreeWidgetItem(parentnode)
                tempitem.setText(0,'%s Results'%name)
                tempitem.setText(1,'%s results in %s set%s'%(self.resultcounts[name],str(storelen),'s' if storelen>1 else ''))
                tempitem.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
                childnodes.append(tempitem)
                
                for k in store:
                    store[k].generateTreeItem(tempitem)

                

class ResultSet(object):
    """defines a group of related findings/results (ex. all findings for a page)"""
    
    def __init__(self, context, pageid, isoverall):
        self.results=list()
        self.stats={}
        self.context=context
        self.pageid=pageid
        self.isoverall=isoverall
        
    def addResult(self,singleresult):
        self.results.append(singleresult)
        self.stats['count']=len(self.results)
    
    def toHTML(self):
        """
        Outputs the results to HTML for viewing in the GUI or reports.
        Override for custom reporting/display functionality
        """
        statsoutput=''
        resultsoutput=''
        
        for key,value in self.stats.items():
            statsoutput+="<em>%s</em>: %s"%(key,value)
        
        resultcounter=1
        for result in self.results:
            resultsoutput+="<hr><h5>Result %s</h5><br>"%(resultcounter,)
            resultcounter+=1
            resultsoutput+=result.toHTML()

        return """
        <h3>%s%s</h3>
        <h4>Analysis Stats</h4>
        %s
        <h4>Results</h4>
        %s
        
        """%(('%s - '%(self.pageid) if self.pageid is not None else ''),
             self.context,
             statsoutput,
             resultsoutput)
        
    def generateTreeItem(self,parentnode):
        tempitem=QTreeWidgetItem(parentnode)
        #context, pageid, isoverall
        if self.isoverall:
            tempitem.setText(0,str(self.context))
        else:
            tempitem.setText(0,'%s: %s'%(self.pageid,self.context))
        tempitem.setText(1,'%s results'%self.numresults)
        tempitem.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
        tempitem.customdata=self
        return tempitem
    
    def generateTreeChildren(self,db,cursor,parentnode):
        #If this tree item came from the db, and we haven't populated it yet, populate it.
        if self.dbgenerated and not self.dbretrieved:
            dbresults=db.analysis_get_singleresults_per_resultset(cursor,self.resultsetid)
            
            for result in dbresults:
                #AnalysisSingleResult_ID, severity, certainty, type, desc,data, spanstart,spanend
                span= None if result[6] is None else (result[6],result[7]) 
                data= ast.literal_eval(result[5])
                tempresult=SingleResult(result[3],result[4],data, span, result[1], result[2])
                self.results.append(tempresult)
            
            dbstats=db.analysis_get_stats_per_resultset(cursor,self.resultsetid)
            for stat in dbstats:
                #AnalysisStat_ID, statName, statValue
                self.stats[stat[1]]=stat[2]
            self.dbretrieved=True
        
        #Data is populated, make the treeitems
        for result in self.results:
            result.generateTreeItem(parentnode)
            
            
class SingleResult(object):
    """defines a a single finding/result"""
    
    #Constants and string translation.
    #TODO: There has to be a better way to do this.
    UNKNOWN=None
    HIGH=1000
    MEDIUM=100
    LOW=10
    
    SEVERITYNAME={None:'Undefined', 1000:'High', 100: 'Medium', 10:'Low'}
    
    #DROP if getMembers no longer needed
    CONSTANTS=('UNKNOWN','HIGH','MEDIUM','LOW','SEVERITYNAME','CONSTANTS')
    
    
    def __init__(self, type, desc, data, span=None, severity=UNKNOWN, certainty=UNKNOWN):
        """type=general type of result.
            desc=friendly description of result
            data=any associated data for the result
            span=the starting and ending string indexes where the problem was found in the response
            severity=One of the constants UNKNOWN, HIGH, MEDIUM, LOW
            certainty=One of the constants UNKNOWN, HIGH,MEDIUM, LOW
            """
        self.severity=severity
        self.certainty=certainty
        self.type=type
        self.desc=desc
        self.data=data
        self.span=span
    
    def toHTML(self):
        
        locationtext=''
        dataoutput=''
        if self.span is not None:
            locationtext='%s to %s'%self.span
        
        # TODO: This is still f'd up.  Look at 7/27 --Seth
        #try: 
        #datalist = ast.literal_eval(self.data)
        dataoutput = '<ol><li>%s</li></ol>' % self.data
        #for k in datalist:
        #   dataoutput += "<li>%s: %s</li>" % (k,datalist[k])
        #dataoutput += '</ol>'
        #except ValueError:
        #    print "ValueError: %s" % self.data
        #except:
        #    print "Error occurred on this data: %s" % self.data

        NiceOutput="""
        <h3>%s</h3>
        <font size="-1">Severity: <b>%s</b> | Certainty: <b>%s | Location: %s</b></font>
        <p><b>Description:</b>
        <ul>
         <li>%s</li>
        </ul>
        <p><b>More Details:</b>
        <li>
         <ul>%s</ul>
        </li>
        """%(self.type,
               SingleResult.SEVERITYNAME[self.severity],
               SingleResult.SEVERITYNAME[self.certainty],
               locationtext,
               self.desc,
               dataoutput)
                
        return NiceOutput
    
    def generateTreeItem(self,parentnode):
        tempitem=QTreeWidgetItem(parentnode)
        tempitem.setText(0,str(self.type))
        tempitem.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
        tempitem.customdata=self
        return tempitem
    
    def generateTreeChildren(self,db,cursor,parentnode):
        return
    
    
    
class AnalysisRun(object):
    def __init__(self,timestamp=None, resultcount=None):
        self.timestamp=timestamp
        self.resultcount=resultcount
        self.analyzerlist=list()
        
    def toHTML(self):
        
        outlist= [a.toHTML() for a in self.analyzerlist]
        
        return "".join(outlist)
    
    def generateTreeItem(self,parentnode):
        tempitem=QTreeWidgetItem(parentnode)
        tempitem.setText(0,str(self.timestamp))
        tempitem.setText(1,str(self.resultcount)+' results')
        tempitem.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
        tempitem.customdata=self
        return tempitem
    
    def generateTreeChildren(self,db,cursor,parentnode):
        for instance in db.analysis_get_instances_per_run(cursor, self.runid):
            instanceid=instance[0]
            tempAR=AnalysisResults()
            tempAR.instanceid=instanceid
            tempAR.dbgenerated=True
            tempAR.dbretrieved=False
            tempAR.setAnalyzerInfo(instance[2],instance[1], instance[3])
            tempAR.numresults=instance[4]
            self.analyzerlist.append(tempAR)
            tempAR.generateTreeItem(parentnode)
