#
# Author: Gregory Fleischer (gfleischer@gmail.com)
#
# Copyright (c) 2011 RAFT Team
#
# This file is part of RAFT.
#
# RAFT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RAFT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RAFT.  If not, see <http://www.gnu.org/licenses/>.
#
from PyQt4.QtCore import Qt, QObject, SIGNAL, QThread, QTimer, QMutex, QUrl
from PyQt4.QtGui import *

import re
from core.database.constants import *
from urllib2 import urlparse

class DomFuzzerThread(QThread):

    UNIQUE_MARKER_BASE = '\'-(/b41ns5xg,)'
    UNIQUE_MARKER = '"'+UNIQUE_MARKER_BASE+'>'
    NUM1 = 857345
    NUM2 = 572912
    STANDARD_TESTS = ['\'"></script><script>alert(%d)</script>' % NUM1, '\'-alert(%d)-\'' % NUM2, UNIQUE_MARKER]

    def __init__(self, framework, dataModel, parent = None):
        QThread.__init__(self, parent)
        self.framework = framework
        self.dataModel = dataModel
        self.qlock = QMutex()
        QObject.connect(self, SIGNAL('quit()'), self.quitHandler)
        QObject.connect(self, SIGNAL('started()'), self.startedHandler)

        self.Data = None
        self.read_cursor = None
        self.write_cursor = None

    def db_attach(self):
        self.Data = self.framework.getDB()
        self.read_cursor = self.Data.allocate_thread_cursor()
        self.write_cursor = self.Data.allocate_thread_cursor()
        self.populateFuzzQueue()

    def db_detach(self):
        self.close_cursor()
        self.Data = None

    def close_cursor(self):
        if self.write_cursor and self.Data:
            self.write_cursor.close()
            self.Data.release_thread_cursor(self.write_cursor)
            self.write_cursor = None
        if self.read_cursor and self.Data:
            self.read_cursor.close()
            self.Data.release_thread_cursor(self.read_cursor)
            self.read_cursor = None

    def run(self):
        QObject.connect(self, SIGNAL('populateFuzzQueue()'), self.do_populateFuzzQueue, Qt.DirectConnection)
        QObject.connect(self, SIGNAL('startFuzzing()'), self.do_startFuzzing, Qt.DirectConnection)
        QObject.connect(self, SIGNAL('stopFuzzing()'), self.do_stopFuzzing, Qt.DirectConnection)
        QObject.connect(self, SIGNAL('fuzzItemFinished()'), self.do_fuzzItemFinished, Qt.DirectConnection)
        self.exec_()

    def quitHandler(self):
        self.framework.debug_log('DomFuzzerThread quit...')
        self.close_cursor()
        self.exit(0)

    def startedHandler(self):
        self.framework.debug_log('DomFuzzerThread started...')
        self.framework.subscribe_database_events(self.db_attach, self.db_detach)
        self.framework.subscribe_populate_domfuzzer_response_id(self.do_populate_domfuzzer_response_id)
        self.framework.subscribe_populate_domfuzzer_response_list(self.do_populate_domfuzzer_response_list)

    def populateFuzzQueue(self):
        QTimer.singleShot(10, self, SIGNAL('populateFuzzQueue()'))

    def startFuzzing(self, fuzz_callback):
        self.fuzz_callback = fuzz_callback
        QTimer.singleShot(10, self, SIGNAL('startFuzzing()'))

    def stopFuzzing(self):
        QTimer.singleShot(10, self, SIGNAL('stopFuzzing()'))

    def fuzzItemFinished(self, fuzz_id, html, messages):
        self.finished_fuzz_item = (fuzz_id, html, messages)
        QTimer.singleShot(10, self, SIGNAL('fuzzItemFinished()'))

    def page_load_finished(self):
        pass ### print('page load finished')
        page = self.__page
        page.setViewPortSize(page.mainFrame().contentsSize())
        image = QImage(page.viewportSize(), QImage.Format_ARGB32)
        painter = QPainter(image)
        page.mainFrame().render(painter)
        pass ### print('finished rendering')

    def do_startFuzzing(self):
        pass ### print('do_startFuzzing')
        self.keep_fuzzing = True
        self.dispatch_next_fuzz_item()

    def do_fuzzItemFinished(self):
        pass ### print('do_fuzzItemFinished', self.finished_fuzz_item)
        fuzz_id, html, message = self.finished_fuzz_item
        self.Data.update_dom_fuzzer_queue_item_status(self.write_cursor, fuzz_id, 'C')
        self.dispatch_next_fuzz_item()

    def dispatch_next_fuzz_item(self):
        if self.keep_fuzzing:
            fuzz_item = self.get_next_fuzz_item()
            if fuzz_item:
                QObject.emit(self.fuzz_callback, SIGNAL('fuzzItemAvailable(int, QString, QUrl)'), fuzz_item[0], fuzz_item[1], fuzz_item[2])
            else:
                QObject.emit(self.fuzz_callback, SIGNAL('fuzzRunFinished()'))

    def get_next_fuzz_item(self):
        self.qlock.lock()
        locked = True
        fuzz_item = None
        try:
            data_item = self.dataModel.popleft_data()
            if not data_item:
                self.keep_fuzzing = False
                return

            fuzz_id = data_item[DomFuzzerQueueTable.ID]
            row = self.Data.read_responses_by_id(self.read_cursor, data_item[DomFuzzerQueueTable.RESPONSE_ID])

            if not row:
                self.framework.log_warning('missing response id: %s' % (response_id))
                return

            responseItems = [m or '' for m in row]
            qurl = QUrl.fromEncoded(str(data_item[DomFuzzerQueueTable.URL]) + str(data_item[DomFuzzerQueueTable.TEST]))
            dataContent = str(responseItems[ResponsesTable.RES_DATA])

            # TODO: store reference
            fuzz_item = (fuzz_id, dataContent, qurl)
            self.qlock.unlock()
            locked = False

        finally:
            if locked:
                self.qlock.unlock()

        return fuzz_item

    def do_stopFuzzing(self):
        pass ### print('do_stopFuzzing')
        self.keep_fuzzing = False

    def do_populate_domfuzzer_response_id(self, response_id):
        pass ### print('do_populate_domfuzzer_response_id', response_id)
        self.generate_fuzz_values(response_id, True)

    def do_populate_domfuzzer_response_list(self, id_list):
        pass ### print('do_populate_domfuzzer_response_list', id_list)

    def generate_fuzz_values(self, response_id, singleRequest):

        row = self.Data.read_responses_by_id(self.read_cursor, response_id)
        if not row:
            self.framework.log_warning('missing response id: %s' % (response_id))
            return
        responseItems = [m or '' for m in row]
        url = responseItems[ResponsesTable.URL]

        fuzz_payloads = self.calculate_fuzz_payloads(url)

        self.qlock.lock()
        try:
            if not singleRequest and self.fuzz_queue.has_key(url):
                # TODO: allow for better rescan in future
                self.framework.log_warning('adding already scanned url [%s] for now' % (url))

            for payload in fuzz_payloads:
                queue_item = [None, response_id, payload[0], payload[1], payload[2], payload[3], 'P']
                rowid = self.Data.add_dom_fuzzer_queue_item(
                    self.write_cursor, queue_item
                    )
                queue_item[0] = rowid
                self.dataModel.append_data([queue_item])
        finally:
            self.qlock.unlock()

    def do_populateFuzzQueue(self):
        self.qlock.lock()
        try:
            rows = []
            for row in self.Data.get_dom_fuzzer_queue_items(self.read_cursor, 'P'):
                rows.append([m or '' for m in row])
            self.dataModel.append_data(rows)
        finally:
            self.qlock.unlock()

    def calculate_fuzz_payloads(self, url):
        payloads = []
        splitted = urlparse.urlsplit(url)
        if not splitted.query:
            if not splitted.fragment:
                payloads.extend(self.calculate_fuzz_payload_tests(url+'?', 'url', ''))
                payloads.extend(self.calculate_fuzz_payload_tests(url+'#', 'url', ''))
            else:
                mangled = urlparse.urlunsplit((splitted.scheme, splitted.netloc, splitted.path, '', ''))
                mangled = '?#' + splitted.fragment
                payloads.extend(self.calculate_fuzz_payload_tests(mangled, 'url', ''))
        if splitted.query:
            payloads.extend(self.calculate_fuzz_payload_tests(url, 'query', splitted.query))
        if splitted.fragment:
            payloads.extend(self.calculate_fuzz_payload_tests(url, 'fragment', splitted.fragment))
        if splitted.query and not splitted.fragment:
            payloads.extend(self.calculate_fuzz_payload_tests(url, 'url', ''))
            mangled = urlparse.urlunsplit((splitted.scheme, splitted.netloc, splitted.path, '', splitted.query))
            payloads.extend(self.calculate_fuzz_payload_tests(mangled, 'fragment', splitted.query))
        return payloads

    def calculate_fuzz_payload_tests(self, url, target, url_field):

        payloads = []

        for test in self.STANDARD_TESTS:
            payloads.append((url, target, '', test))

        if url_field:
            if ';' in url_field:
                delim = ';'
            else:
                delim = '&'

            pairs = url_field.split(delim)
            for test in self.STANDARD_TESTS:
                for offset in range(0, len(pairs)):
                    values = pairs[offset]
                    if '=' in values:
                        name, value = values.split('=', 1)
                    else:
                        name, value = values, ''
                    payloads.append((url, target, name, value+test))

        return payloads
