#
# Author: Seth Law
#
# Copyright (c) 2011 RAFT Team
#
# This file is part of RAFT.
#
# RAFT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# RAFT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RAFT.  If not, see <http://www.gnu.org/licenses/>.
#
import re

from ..AbstractAnalyzer import AbstractAnalyzer


class FindInsecureCookies(AbstractAnalyzer):
    
    #Class variables shared across all instances
    SetCookieRegex = re.compile("set-cookie\:\s(\S*?=\S*?);\s(.*)",re.I)
    SecureRegex = re.compile("secure",re.I)
    HttpOnlyRegex = re.compile("httponly",re.I)
    HttpsRegex = re.compile("https://",re.I)
    notSecureText = "If the 'secure' flag is included in the HTTP set-cookie response header, the cookie will only be sent with a request when the browser is communicating using HTTPS."
    notSecure = {}
    notHttpOnlyText = "If the HttpOnly flag (optional) is included in the HTTP response header, the cookie cannot be accessed through client side script (again if the browser supports this flag). As a result, even if a cross-site scripting (XSS) flaw exists, and a user accidentally accesses a link that exploits this flaw, the browser (primarily Internet Explorer) will not reveal the cookie to a third party. - https://www.owasp.org/index.php/HttpOnly"
    notHttpOnly = {}
    cookies = {}
    
    def __init__(self):
        self.desc="Identifies Cookies that are set using insecure parameters."
        self.friendlyname="Find Insecure Cookies"
    
    def analyzeTransaction(self, target, results):
        responseHeaders=target['response']['headers']
        host = target['request']['host']
        url = target['url']
        
            
        for found in FindInsecureCookies.SetCookieRegex.finditer(responseHeaders):
                
                cookie = found.group(1)
                params = found.group(2)
                (name,value) = re.split("=",cookie)
                
                try:
                    self.cookies[host]
                except:
                    self.cookies[host] = {}
                try:
                    self.cookies[host][name]
                except:
                    self.cookies[host][name] = []
                self.cookies[host][name].append(value)
                
                if ((FindInsecureCookies.SecureRegex.search(params) == None) and (self.HttpsRegex.search(url))):
                    #print "=> FAIL (secure)"
                    try:
                        self.notSecure[host]
                    except:
                        self.notSecure[host] = {}
                    
                    try:
                        self.notSecure[host][name]
                    except:
                        self.notSecure[host][name] = 0
                    self.notSecure[host][name] += 1
                    results.addPageResult(pageid=target['id'], 
                                  url=target['url'],
                                  type="Cookie not marked as 'secure'",
                                  desc=self.notSecureText,
                                  data={'Cookie':name},
                                  span=found.span())
                    
                if (FindInsecureCookies.HttpOnlyRegex.search(params) == None):
                    try:
                        self.notHttpOnly[host]
                    except:
                        self.notHttpOnly[host] = {}
                        
                    try:
                        self.notHttpOnly[host][name]
                    except:
                        self.notHttpOnly[host][name] = 0
                    self.notHttpOnly[host][name] += 1
                    results.addPageResult(pageid=target['id'], 
                                  url=target['url'],
                                  type="Cookie not marked as 'httponly'",
                                  desc=self.notHttpOnlyText,
                                  data={'Cookie':name},
                                  span=found.span())

    def postanalysis(self,results):
        outputSecureHeader = 'Cookies not marked as "secure"'
        outputSecureValue = ""
        for h in self.notSecure.iterkeys():
            outputSecureValue += " %s (" % h
            first = True
            for v in self.notSecure[h].iterkeys():
                if not first:
                    outputSecureValue += ", "
                else:
                    first = False
                outputSecureValue += "%s" % v
            outputSecureValue += ")"
            
            results.addOverallResult(type=self.friendlyname,
                                     desc=self.desc,
                                     data={outputSecureHeader:outputSecureValue},
                                     span=None,
                                     certainty=None,
                                     context=h
                                     )

        outputHttpHeader = 'Cookies not marked as "httponly"'
        outputHttpValue = ""
        for h in self.notHttpOnly.iterkeys():
            outputHttpValue += " %s (" % h
            first = True
            for v in self.notHttpOnly[h].iterkeys():
                if not first:
                    outputHttpValue += ", "
                else:
                    first = False
                outputHttpValue += "%s" % v
            outputHttpValue += ")"
               
            results.addOverallResult(type=self.friendlyname,
                                     desc=self.desc,
                                     data={outputHttpHeader:outputHttpValue},
                                     span=None,
                                     certainty=None,
                                     context=h
                                     )