#
# request runner
#
# Authors:
#          Gregory Fleischer (gfleischer@gmail.com)
#
# Copyright (c) 2011 RAFT Team
#
# This file is part of RAFT.
#
# RAFT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RAFT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RAFT.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt4.QtCore import (Qt, QObject, SIGNAL, QMutex)

from core.network.StandardNetworkAccessManager import StandardNetworkAccessManager
from core.network.InMemoryCookieJar import InMemoryCookieJar
from core.network.NetworkRequester import NetworkRequester
from core.fuzzer.RequestInstance import RequestInstance
from core.fuzzer.SequenceManager import SequenceManager

import collections
import uuid

class RequestRunner(QObject):

    def __init__(self, framework, parent = None):
        QObject.__init__(self, parent)

        self.framework = framework

        self.request_queue = None

        self.qlock = QMutex()
        self.cookieJar = None
        self.networkRequester = None
        self.networkAccessManager = None
        self.sequenceManager = None

        self.request_queue = collections.deque()
        self.request_context = {}

        self.Data = None
        self.cursor = None
        self.framework.subscribe_database_events(self.db_attach, self.db_detach)

        self.max_concurrent = 10

    def db_attach(self):
        self.Data = self.framework.getDB()
        self.cursor = self.Data.allocate_thread_cursor()

    def db_detach(self):
        self.close_cursor()
        self.Data = None

    def close_cursor(self):
        if self.cursor and self.Data:
            self.cursor.close()
            self.Data.release_thread_cursor(self.cursor)
            self.cursor = None

    def setup(self, callback, use_global_cookies = True, sequenceId = None):

        QObject.connect(self, SIGNAL('requestFinished(int, QString)'), callback, Qt.DirectConnection)

        self.sequence_enabled = False
        self.sequence_needed = False
        if sequenceId:
            self.sequenceManager = SequenceManager(self.framework, sequenceId, self)
            self.sequence_enabled = True
            self.sequence_needed = True

        if use_global_cookies:
            self.cookieJar = self.framework.get_global_cookie_jar()
        else:
            self.cookieJar = InMemoryCookieJar(self.framework, self)

        self.networkAccessManager = StandardNetworkAccessManager(self.framework, self.cookieJar)
        self.networkRequester = NetworkRequester(self.framework, self.networkAccessManager, self.response_received, self)

        self.inflight_list = {}

    def queue_request(self, method, url, headers, body, context = ''):
        if not context:
            context = uuid.uuid4().hex

        request = RequestInstance(method, url, headers, body, context, self)

        do_process = False

        self.qlock.lock()
        try:

            self.request_context[request.context] = request
            self.request_queue.append(request)
            if len(self.inflight_list) < self.max_concurrent:
                do_process = True

        finally:
            self.qlock.unlock()

        if do_process:
            self.process_next()

        return request

    def response_received(self, response):
        #print(response.response_id, response.context)
        do_process = False
        send_signal = False
        user_request = False
        original_request = None
        self.qlock.lock()
        try:
            if not self.inflight_list.has_key(response.context):
                raise Exception('unexpected response; context=%s' % (response.context))
            request = self.inflight_list.pop(response.context)

            if self.request_context.has_key(response.context):
                # user request
                if not self.sequence_enabled:
                    send_signal = True
                    self.request_context.pop(response.context)
                else:
                    original_request = self.request_context[response.context]
                    send_signal = True
                    need_sequence, run_again = self.sequenceManager.analyze_response(response)
                    if run_again:
                        self.request_queue.appendleft(original_request)
                        send_signal = False
                    else:
                        self.request_context.pop(response.context)
                    if need_sequence:
                        self.sequence_needed = True

            if len(self.inflight_list) < self.max_concurrent:
                do_process = True

        finally:
            self.qlock.unlock()

        if send_signal:
            self.emit(SIGNAL('requestFinished(int, QString)'), response.response_id, response.context)

        if do_process:
            self.process_next()

    def process_next(self):
        is_locked = False
        try:
            keep_looping = True
            while keep_looping:

                self.qlock.lock()
                is_locked = True

                if len(self.request_queue) == 0:
                    keep_looping = False
                    break

                if self.sequence_enabled and self.sequence_needed:
                    self.setup_sequence_items()
                    self.sequence_needed = False

                request = self.request_queue.popleft()
#                print('process_next', request)
#                print(self.request_queue)
#                print(self.inflight_list)
                if self.request_context.has_key(request.context):
                    # user request - protect cookies
                    self.cookieJar.set_read_only(True)
                self.inflight_list[request.context] = self.networkRequester.send(
                    request.method, request.url, request.headers, request.body, request.context
                    )
                if len(self.inflight_list) >= self.max_concurrent:
                    keep_looping = False

                self.qlock.unlock()
                is_locked = False

        finally:
            if is_locked:
                self.qlock.unlock()

    def setup_sequence_items(self):
        # clear cookies
        self.cookieJar.set_read_only(False)
        self.cookieJar.clear_cookies()
        # this function must be called with lock
        request_list = self.sequenceManager.get_request_list()
        for i in range(len(request_list), 0, -1):
            self.request_queue.appendleft(request_list[i-1])
